/*
 * NXTway-GS model
 * Plant:      continuous time, linear/non-linear
 * Controller: continuous time, linear/saturated
 *             discrete time, linear/saturated
 * All 8 combinations of plant and controller are instantiated below.
 *
 * Based on a Simulink model by Yorihisa Yamamoto, The MathWorks, Inc.
 * http://www.mathworks.com/matlabcentral/fileexchange/19147-nxtway-gs-self-balancing-two-wheeled-robot-controller-design
 *
 * Modelica model by Thomas Heinz, Robert Bosch GmbH
 *
 */

model System
  constant Real pi = 3.141592653589793;

model Plant
  constant Real g = 9.81;
  constant Real m = 0.03;
  constant Real R = 0.04;
  constant Real Jw = (m * R^2) / 2.0;
  constant Real M = 0.6;
  constant Real W = 0.14;
  constant Real D = 0.04;
  constant Real H = 0.144;
  constant Real L = H/2;
  constant Real Jpsi = (M * L^2) / 3.0;
  constant Real Jphi = (M * W^2 + D^2) / 12.0;
  constant Real Jm = 0.00001;
  constant Real Rm = 6.69;
  constant Real Kb = 0.468;
  constant Real Kt = 0.317;
  constant Real n = 1.0;
  constant Real fm = 0.0022;
  constant Real fw = 0.0;
  constant Real _alpha = (n * Kt) / Rm;
  constant Real _beta = (n * Kt * Kb) / Rm + fm;
  constant Real I_ = (m * W^2) / 2.0 + Jphi + (W^2 * (Jw + n^2 * Jm)) / (2.0 * R^2);
  constant Real J = (W^2 * (_beta + fw)) / (2.0 * R^2);
  constant Real K = (W * _alpha) / (2.0 * R);
  constant Real battery = 8000.0;
  constant Real gyro_offset = 600.0;
  constant Real[3, 6] Cp =
    {{0, 0, 1, 0, 0, 0},
     {0, 180/pi, 0, -180/pi, 0, -(90*W)/(R*pi)},
     {0, 180/pi, 0, -180/pi, 0, (90*W)/(R*pi)}};
  constant Real[2, 2] Dp =
    {{0.00001089*battery - 0.00625, 0},
     {0, 0.00001089*battery - 0.00625}};

model linear
  constant Real[6, 6] Ap =
    {{-(2.0*_beta*(2.0*Jm*n^2 - L*M*R) - 2.0*(_beta + fw)*(M*L^2 + 2.0*Jm*n^2 + Jpsi))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw)), 0, -(2.0*_beta*(M*L^2 + M*R*L + Jpsi))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw)), -(L*M*g*(2.0*Jm*n^2 - L*M*R))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw)), 0, 0}, 
     {1, 0, 0, 0, 0, 0},
     {-(2.0*_beta*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw) - 2.0*(_beta + fw)*(2.0*Jm*n^2 - L*M*R))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw)), 0, (2.0*_beta*((M + 2.0*m)*R^2 + L*M*R + 2.0*Jw))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw)), -(L*M*g*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw)), 0, 0},
     {0, 0, 1, 0, 0, 0},
     {0, 0, 0, 0, -J/I_, 0},
     {0, 0, 0, 0, 1, 0}};
  constant Real[6, 2] Bp =
    {{-(_alpha*(M*L^2 + M*R*L + Jpsi))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw)), -(_alpha*(M*L^2 + M*R*L + Jpsi))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw))},
     {0, 0},
     {(_alpha*((M + 2.0*m)*R^2 + L*M*R + 2.0*Jw))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw)), (_alpha*((M + 2.0*m)*R^2 + L*M*R + 2.0*Jw))/((2.0*Jm*n^2 - L*M*R)^2 - (M*L^2 + 2.0*Jm*n^2 + Jpsi)*((M + 2.0*m)*R^2 + 2.0*Jm*n^2 + 2.0*Jw))},
     {0, 0},
     {-K/I_, K/I_},
     {0, 0}};

  Real[6] x;
  Real[2] u;
  input Real[2] u_;
  output Real[3] y;

  equation
    der(x) = Ap * x + Bp * u;
    u = Dp * u_;
    y = Cp * x;
end linear;

model nonlinear
  Real[6] x;
  Real[2] u;
  input Real[2] u_;
  output Real[3] y;

  equation
    der(x[2]) = x[1];
    der(x[4]) = x[3];
    der(x[6]) = x[5];
    ((2 * m + M) * R^2 + 2 * Jw + 2 * n^2 * Jm) * der(x[1]) + (M * L * R * cos(x[4]) - 2 * n^2 * Jm) * der(x[3]) - M * L * R * x[3]^2 * sin(x[4]) = _alpha * (u[1] + u[2]) - 2 * (_beta + fw) * x[1] + 2 * _beta * x[3];
    (M * L * R * cos(x[4]) - 2 * n^2 * Jm) * der(x[1]) + (M * L^2 + Jpsi + 2 * n^2 * Jm) * der(x[3]) - M * g * L * sin(x[4]) - M * L^2 * x[5]^2 * sin(x[4]) * cos(x[4]) = -_alpha * (u[1] + u[2]) + 2 * _beta * x[1] - 2 * _beta * x[3];
    (0.5 * m * W^2 + Jphi + W^2 / (2 * R^2) * (Jw + n^2 * Jm) + M * L^2 * (sin(x[4]))^2) * der(x[5]) + 2 * M * L^2 * x[3] * x[5] * sin(x[4]) * cos(x[4]) = W / (2 * R) * _alpha * (u[2] - u[1]) - W^2 / (2 * R^2) * (_beta + fw) * x[5];
    u = Dp * u_;
    y = Cp * x;
end nonlinear;

end Plant;


model Controller
  constant Real CMD_MAX = 100.0;
  constant Real EXEC_PERIOD = 0.00400000019;
  constant Real K_THETADOT = 7.5;
  constant Real K_PHIDOT = 25.0;
  constant Real K_I = -0.44721;
  constant Real K_F0 = -0.834434;
  constant Real K_F1 = -38.1749;
  constant Real K_F2 = -1.0985;
  constant Real K_F3 = -3.55477;
  constant Real A_R = 0.996;
  constant Real A_D = 0.8;
  constant Real DEG2RAD = 0.01745329238;
  constant Real BATTERY_GAIN = 0.001089;
  constant Real BATTERY_OFFSET = 0.625;
  constant Real[2, 5] Cc =
    {{-(100.0*K_I)/(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery), -(100.0*K_F0)/(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery), -(100.0*A_R*K_F2)/(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery), (100.0*(K_F2 - 1.0*A_D*K_F2 + EXEC_PERIOD*K_F0 + EXEC_PERIOD*K_F1))/(EXEC_PERIOD*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), (100.0*K_F2*(A_D - 1.0))/(EXEC_PERIOD*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery))},
     {-(100.0*K_I)/(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery), -(100.0*K_F0)/(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery), -(100.0*A_R*K_F2)/(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery), (100.0*(K_F2 - 1.0*A_D*K_F2 + EXEC_PERIOD*K_F0 + EXEC_PERIOD*K_F1))/(EXEC_PERIOD*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), (100.0*K_F2*(A_D - 1.0))/(EXEC_PERIOD*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery))}};
  constant Real[2, 5] Dc =
    {{(18000.0*DEG2RAD*K_F3)/(pi*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), (50.0*DEG2RAD*(K_F2 - 1.0*A_D*K_F2 + EXEC_PERIOD*K_F0))/(EXEC_PERIOD*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), (50.0*DEG2RAD*(K_F2 - 1.0*A_D*K_F2 + EXEC_PERIOD*K_F0))/(EXEC_PERIOD*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), (100.0*K_F2*K_THETADOT*(A_R - 1.0))/(CMD_MAX*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), K_PHIDOT/CMD_MAX},
     {(18000.0*DEG2RAD*K_F3)/(pi*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), (50.0*DEG2RAD*(K_F2 - 1.0*A_D*K_F2 + EXEC_PERIOD*K_F0))/(EXEC_PERIOD*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), (50.0*DEG2RAD*(K_F2 - 1.0*A_D*K_F2 + EXEC_PERIOD*K_F0))/(EXEC_PERIOD*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), (100.0*K_F2*K_THETADOT*(A_R - 1.0))/(CMD_MAX*(BATTERY_OFFSET - 1.0*BATTERY_GAIN*Plant.battery)), -(1.0*K_PHIDOT)/CMD_MAX}};

model continuous_t
  constant Real[5, 5] Ac =
    {{0, 1.0, 0, -1.0, 0},
     {0, 0, A_R, 0, 0},
     {0, 0, (A_R - 1.0)/EXEC_PERIOD, 0, 0},
     {0, 0, 0, 0, 0},
     {0, 0, 0, -(A_D - 1)/EXEC_PERIOD, (A_D - 1.0)/EXEC_PERIOD}};
  constant Real[5, 5] Bc =
    {{0, -0.5*DEG2RAD, -0.5*DEG2RAD, 0, 0},
     {0, 0, 0, -(K_THETADOT*(A_R - 1))/CMD_MAX, 0},
     {0, 0, 0, -(K_THETADOT*(A_R - 1))/(CMD_MAX*EXEC_PERIOD), 0},
     {(180.0*DEG2RAD)/pi, 0, 0, 0, 0},
     {0, -(0.5*DEG2RAD*(A_D - 1))/EXEC_PERIOD, -(0.5*DEG2RAD*(A_D - 1))/EXEC_PERIOD, 0, 0}};

model linear
  Real[5] x;
  Real[5] u;
  input Real[3] u_;
  input Real[2] u__;
  output Real[2] y;

  equation
    der(x) = Ac * x + Bc * u;
    u = vector([u_; u__]);
    y = Cc * x + Dc * u;
end linear;

model saturate
  Real[5] x;
  Real[5] u;
  Real[2] y_;
  input Real[3] u_;
  input Real[2] u__;
  output Real[2] y;

  equation
    der(x) = Ac * x + Bc * u;
    u = vector([u_; u__]);
    y_ = Cc * x + Dc * u;
    for i in 1:2 loop
      y[i] = if y_[i] < -CMD_MAX then -CMD_MAX elseif y_[i] > CMD_MAX then CMD_MAX else y_[i];
    end for;
end saturate;

end continuous_t;

model discrete_t
  constant Real[5, 5] Ac =
    {{1, EXEC_PERIOD, 0, -1.0*EXEC_PERIOD, 0},
     {0, 1, A_R*EXEC_PERIOD, 0, 0},
     {0, 0, A_R, 0, 0},
     {0, 0, 0, 1, 0},
     {0, 0, 0, 1.0 - 1.0*A_D, A_D}};
  constant Real[5, 5] Bc =
    {{0, -0.5*DEG2RAD*EXEC_PERIOD, -0.5*DEG2RAD*EXEC_PERIOD, 0, 0},
     {0, 0, 0, -(EXEC_PERIOD*K_THETADOT*(A_R - 1))/CMD_MAX, 0},
     {0, 0, 0, -(K_THETADOT*(A_R - 1))/CMD_MAX, 0},
     {(180.0*DEG2RAD*EXEC_PERIOD)/pi, 0, 0, 0, 0},
     {0, -0.5*DEG2RAD*(A_D - 1.0), -0.5*DEG2RAD*(A_D - 1.0), 0, 0}};

model linear
  discrete Real[5] x;
  Real[5] u;
  input Real[3] u_;
  input Real[2] u__;
  discrete output Real[2] y;

  equation
    u = vector([u_; u__]);
    when sample(0, EXEC_PERIOD) then
      x = Ac * pre(x) + Bc * u;
      y = Cc * pre(x) + Dc * u;
    end when;
end linear;

model saturate
  discrete Real[5] x;
  Real[5] u;
  discrete Real[2] y_;
  input Real[3] u_;
  input Real[2] u__;
  discrete output Real[2] y;

  equation
    u = vector([u_; u__]);
    when sample(0, EXEC_PERIOD) then
      x = Ac * pre(x) + Bc * u;
      y_ = Cc * pre(x) + Dc * u;
      for i in 1:2 loop
        y[i] = if y_[i] < -CMD_MAX then -CMD_MAX elseif y_[i] > CMD_MAX then CMD_MAX else y_[i];
      end for;
    end when;

  initial equation
    for i in 1:2 loop
      y_[i] = 0;
    end for;
end saturate;

end discrete_t;

end Controller;


Real theta_dot_ref, phi_dot_ref;
Plant.linear pl_ccl(u_ = ccl_pl.y);
Plant.linear pl_ccs(u_ = ccs_pl.y);
Plant.linear pl_cdl(u_ = cdl_pl.y);
Plant.linear pl_cds(u_ = cds_pl.y);
Plant.nonlinear pn_ccl(u_ = ccl_pn.y);
Plant.nonlinear pn_ccs(u_ = ccs_pn.y);
Plant.nonlinear pn_cdl(u_ = cdl_pn.y);
Plant.nonlinear pn_cds(u_ = cds_pn.y);

Controller.continuous_t.linear ccl_pl(u_ = pl_ccl.y, u__ = vector([theta_dot_ref; phi_dot_ref]));
Controller.continuous_t.linear ccl_pn(u_ = pn_ccl.y, u__ = vector([theta_dot_ref; phi_dot_ref]));
Controller.continuous_t.saturate ccs_pl(u_ = pl_ccs.y, u__ = vector([theta_dot_ref; phi_dot_ref]));
Controller.continuous_t.saturate ccs_pn(u_ = pn_ccs.y, u__ = vector([theta_dot_ref; phi_dot_ref]));
Controller.discrete_t.linear cdl_pl(u_ = pl_cdl.y, u__ = vector([theta_dot_ref; phi_dot_ref]));
Controller.discrete_t.linear cdl_pn(u_ = pn_cdl.y, u__ = vector([theta_dot_ref; phi_dot_ref]));
Controller.discrete_t.saturate cds_pl(u_ = pl_cds.y, u__ = vector([theta_dot_ref; phi_dot_ref]));
Controller.discrete_t.saturate cds_pn(u_ = pn_cds.y, u__ = vector([theta_dot_ref; phi_dot_ref]));

equation
  theta_dot_ref = if time >= 70 then -100.0 elseif time >= 40 then 100.0 else 0;
  phi_dot_ref = if time >= 70 then 100.0 elseif time >= 40 then -100.0 else 0;

initial equation
  for i in {1, 2, 3, 5, 6} loop
    pl_ccl.x[i] = 0;
    pl_ccs.x[i] = 0;
    pl_cdl.x[i] = 0;
    pl_cds.x[i] = 0;
    pn_ccl.x[i] = 0;
    pn_ccs.x[i] = 0;
    pn_cdl.x[i] = 0;
    pn_cds.x[i] = 0;
  end for;
  pl_ccl.x[4] = 0.15;
  pl_ccs.x[4] = 0.15;
  pl_cdl.x[4] = 0.15;
  pl_cds.x[4] = 0.15;
  pn_ccl.x[4] = 0.15;
  pn_ccs.x[4] = 0.15;
  pn_cdl.x[4] = 0.15;
  pn_cds.x[4] = 0.15;

  for i in 1:5 loop
    ccl_pl.x[i] = 0;
    ccl_pn.x[i] = 0;
    ccs_pl.x[i] = 0;
    ccs_pn.x[i] = 0;
    cdl_pl.x[i] = 0;
    cdl_pn.x[i] = 0;
    cds_pl.x[i] = 0;
    cds_pn.x[i] = 0;
  end for;
end System;
